/*
** resolver.c
**
** name resolver library
**
*/

#ifdef _WIN32
#include <win32config.h>
#elif defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#include <stdio.h>

#ifndef __USE_MISC
# define __USE_MISC
#endif

#ifndef __USE_GNU
# define __USE_GNU
#endif

#ifndef __USE_BSD
# define __USE_BSD
#endif

#ifndef __EXTENSIONS__
# define __EXTENSIONS__
#endif

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif

#ifndef _THREAD_SAFE
# define _THREAD_SAFE
#endif

#ifndef _REENTRANT
# define _REENTRANT
#endif

/* This for freebsd (needed on 3.2 at least) */
#ifdef SOMEBSD
# ifndef _POSIX_VERSION
# define _POSIX_VERSION 199309L
# endif
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>

#ifndef _WIN32
#include <pthread.h>
#endif

#include "resolver.h"

/* internal function */

static void _lock_resolver(void);
static void _unlock_resolver(void);
static char *_lookup(const char *what, char *buff, int len);
static int _isip(const char *what);

/* internal data */

#ifdef _WIN32
#define mutex_t CRITICAL_SECTION
#else
#define mutex_t pthread_mutex_t
#endif

mutex_t _resolver_mutex;

static int _isip(const char *what)
{
	struct in_addr inp;

	return inet_aton(what, &inp);
}

char *resolver_getname(const char *ip, char *buff, int len)
{
	if (!_isip(ip)) {
		strncpy(buff, ip, len);
		return buff;
	}

	return _lookup(ip, buff, len);
}

char *resolver_getip(const char *name, char *buff, int len)
{
	if (_isip(name)) {
		strncpy(buff, name, len);
		return buff;
	}

	return _lookup(name, buff, len);
}

static char *_lookup(const char *what, char *buff, int len)
{
	struct in_addr inp;
	struct hostent *host;
	char *temp;

	/* do a little sanity checking */
	if (what == NULL || buff == NULL || len <= 0)
		return NULL;

	/* gotta lock calls for now, since gethostbyname and such aren't threadsafe */
	_lock_resolver();

	if (_isip(what)) {
		host = gethostbyaddr((char *)&inp, sizeof(struct in_addr), AF_INET);
		if (host == NULL) {
			buff = NULL;
		} else {
			strncpy(buff, host->h_name, len);
		}
	} else {
		host = gethostbyname(what);
		if (host == NULL) {
			buff = NULL;
		} else {
			temp = inet_ntoa(*(struct in_addr *)host->h_addr);
			strncpy(buff, temp, len);
		}
	}

	_unlock_resolver();

	return buff;
}

void resolver_initialize()
{
	/* initialize the lib if we havne't done so already */

#ifndef _WIN32
	pthread_mutex_init(&_resolver_mutex, NULL);
#else
	InitializeCriticalSection(&_resolver_mutex);
#endif

	/* keep dns connects (TCP) open */
	sethostent(1);
}

static void _lock_resolver()
{
#ifndef _WIN32
	pthread_mutex_lock(&_resolver_mutex);
#else
	EnterCriticalSection(&_resolver_mutex);
#endif
}

static void _unlock_resolver()
{
#ifndef _WIN32
	pthread_mutex_unlock(&_resolver_mutex);
#else
	LeaveCriticalSection(&_resolver_mutex);
#endif	
}

#if defined(_WIN32) || !defined(HAVE_INET_ATON)
/* 
 * guess this isn't part of win32. weird! -- ajs 3/99 
 * Assert Class: 2 
 */
int inet_aton(const char *s, struct in_addr *a)
{
	int lsb, b2, b3, msb;
	
	if (sscanf(s, "%d.%d.%d.%d", &lsb, &b2, &b3, &msb) < 4) {
		return 0;
	}
#if defined(HAVE_INET_ADDR)
	a->s_addr = inet_addr(s);
	
	if (a->s_addr == -1) /* que? */
		return 0;
	return 1;
#else

	a->s_addr = lsb + (b2 << 8) + (b3 << 16) + (msb << 24);
	
	return 1;
#endif
}
#endif
