/* thread.h
 * - Thread Abstraction Function Headers
 *
 * Copyright (c) 1999, 2000 the icecast team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#ifndef __THREAD_H__
#define __THREAD_H__

#ifdef _WIN32
typedef DWORD sys_thread_t;
typedef CRITICAL_SECTION sys_mutex_t;
#else
#include <pthread.h>
typedef pthread_t sys_thread_t;
typedef pthread_mutex_t sys_mutex_t;
#endif

typedef struct thread_tag {
	/* the local id for the thread, and it's name */
	long thread_id;
	char *name;

	/* the time the thread was created */
	time_t create_time;
	
	/* the file and line which created this thread */
	char *file;
	int line;

	/* the system specific thread */
	sys_thread_t sys_thread;
} thread_t;

typedef struct mutex_tag {
	/* the local id and name of the mutex */
	long mutex_id;
	char *name;

	/* the thread which is currently locking this mutex */
	long thread_id;

	/* the file and line where the mutex was locked */
	char *file;
	int line;	

	/* the system specific mutex */
	sys_mutex_t sys_mutex;
} mutex_t;


#define thread_create(n,x,y) thread_create_c(n,x,y,__LINE__,__FILE__)
#define thread_create_mutex(x) thread_create_mutex_c(x,__LINE__,__FILE__)
#define thread_lock_mutex(x) thread_mutex_lock_c(x,__LINE__,__FILE__)
#define thread_unlock_mutex(x) thread_mutex_unlock_c(x,__LINE__,__FILE__)
#define thread_exit(x) thread_exit_c(x,__LINE__,__FILE__)

#define MUTEX_STATE_NOTLOCKED -1
#define MUTEX_STATE_NEVERLOCKED -2
#define MUTEX_STATE_UNINIT -3

/* init/shutdown of the library */
void thread_initialize(void);
void thread_initialize_with_log_id(int log_id);
void thread_shutdown(void);

/* creation, destruction, locking, and unlocking */
long thread_create_c(char *name, void *(*start_routine)(void *), void *arg, int line, char *file);
void thread_create_mutex_c(mutex_t *mutex, int line, char *file);
void thread_mutex_lock_c(mutex_t *mutex, int line, char *file);
void thread_mutex_unlock_c(mutex_t *mutex, int line, char *file);
void thread_destroy_mutex(mutex_t *mutex);
void thread_exit_c(int val, int line, char *file);

/* sleeping */
void thread_sleep(unsigned long len);

/* for using library functions which aren't threadsafe */
void thread_library_lock(void);
void thread_library_unlock(void);
#define PROTECT_CODE(code) { thread_library_lock(); code; thread_library_unlock(); }

/* thread information functions */
thread_t *thread_self(void);

/* renames current thread */
void thread_rename(const char *name);

#endif  /* __THREAD_H__ */
